""" Main plotting tools."""
import numpy
import matplotlib.pyplot as plt
from matplotlib.gridspec import (GridSpec as GS,
                                 GridSpecFromSubplotSpec as sGS)
from rhinestone.widgets import (Widget, Slider, Button,
                                RadioButtons, TrianglePlot, CheckButtons)
from anesthetic.anesthetic import NestedSamples


class Higson(Widget):
    """ Higson plot as shown in https://arxiv.org/abs/1703.09701 .

    Attributes:
        curve (matplotlib.lines.Line2D):
            points currently plotted as a curve.

        point (matplotlib.lines.Line2D):
            large indicator point currently plotted on the curve.
    """
    def __init__(self, fig, gridspec):
        super(Higson, self).__init__(fig, gridspec)
        self.ax.set_yticks([])
        self.ax.set_ylim(-0.1, 1.1)
        self.ax.set_ylabel(r'$LX$')
        self.ax.set_xlabel(r'$\log X$')

        self.curve, = self.ax.plot([None], [None], 'k-')
        self.point, = self.ax.plot([None], [None], 'ko')

    def update(self, logX, LX, i):
        """ Update the line and the point in the higson plot.

        Args:
            logX (array):
                log-volume compression values to plot

            LX (array):
                Likelihood * volume compression

            i (int):
                Current location of higson point
        """
        self.point.set_xdata(logX[i])
        self.point.set_ydata(LX[i])
        self.curve.set_xdata(logX)
        self.curve.set_ydata(LX)

    def reset_range(self):
        """ Reset the ranges of the higson plot. """
        xdata = self.curve.get_xdata()
        self.ax.set_xlim(max(xdata), min(xdata))


class Evolution(Slider):
    """ Slider controlling the evolution stage of the live points."""
    def __init__(self, fig, gridspec, action, valmax):
        super(Evolution, self).__init__(fig, gridspec, action, '',
                                        0, valmax, 0, 'horizontal')
        self.slider.valtext.set_horizontalalignment('right')
        self.slider.valtext.set_position((0.98, 0.5))

    def __call__(self):
        """ Return the current iteration as an integer. """
        mx = self.slider.valmax-1
        val = int(super(Evolution, self).__call__())
        return min(val, mx)

    def set_text(self, logL, n):
        """ Set the text at end of slider.

        Args:
            logL (float):
                Current loglikelihood of evolution stage

            n (float):
                Current number of live points of evolution stage
        """
        text = r'$\log L$: %.6g, $n_\mathrm{live}$: %i' % (logL, n)
        return super(Evolution, self).set_text(text)


class Temperature(Slider):
    """ Logarithmic slider controlling temperature of the posterior points."""
    def __init__(self, fig, gridspec, action):
        super(Temperature, self).__init__(fig, gridspec, action, r'$kT$',
                                          -1, 5, 0, 'vertical')

    def __call__(self):
        """ Return the current temperature. """
        return 10**super(Temperature, self).__call__()

    def set_text(self, kT):
        """ Set the text at end of slider.

        Args:
            kT (float):
                Current temperature of posterior points stage
        """
        text = r'%.2g' % kT
        return super(Temperature, self).set_text(text)


class RunPlotter(object):
    """ Construct a control panel of information on a nested sampling run.

    Args:
        root (str):
            The root string for the chains files to be used.

    Attributes:
        run (anesthetic.anesthetic.NestedSamples):
            Object for extracting nested sampling data from chains files.

        fig (matplotlib.figure.Figure):
            Reference to the underlying figure

        triangle (rhinestone.widgets.TrianglePlot):
            Corner plot of live or posterior samples.

        temperature (rhinestone.plot.Temperature):
            Slider selecting the posterior temperature.

        evolution (rhinestone.plot.Evolution):
            Slider selecting the live iteration.

        higson (rhinestone.plot.Higson):
            Higson plot of posterior weights.

        reset (rhinestone.widgets.Button):
            Button that resets the parameter ranges.

        reload (rhinestone.widgets.Button):
            Button that reloads the files.

        type (rhinestone.widgets.RadioButtons):
            Radio buttons that selects whether to plot live or posteriors.

        param_choice (rhinestone.widgets.CheckButtons):
            Checkbox that selects which parameters to plot.
    """
    def __init__(self, root, labels=None):
        self.run = NestedSamples.read(root)
        if labels:
            self.labels = numpy.array(labels)
        else:
            self.labels = numpy.array(self.run.paramnames[:10])

        self.fig = plt.figure()
        self._set_up()
        self.redraw(None)

    def _set_up(self):
        """ Draw the control panel.

        We implement the control panel using sequential recursive gridspecs.

        +----------------------------------------------+------+
        |                                              |      |
        |                  gs0[0]                      |gs0[1]|
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        +-------------------------+-----------------+--+------+
        |                         |    gs11[0]      |         |
        |         gs10[0]         +-----------------+  gs1[2] |
        +-------------------------+    gs11[1]      |         |
        |                         +-----------------+         |
        |         gs10[1]         |    gs11[2]      |         |
        +-------------------------+-----------------+---------+

        These variable names are included in the __init__ function, and are
        named with an intuitive Huffman coding.
        """

        gs = GS(2, 1, height_ratios=[3, 1])
        gs0 = sGS(1, 2, width_ratios=[19, 1], subplot_spec=gs[0])
        gs1 = sGS(1, 3, width_ratios=[4, 1, 1], subplot_spec=gs[1])
        gs10 = sGS(2, 1, height_ratios=[1, 4], subplot_spec=gs1[0])
        gs11 = sGS(3, 1, height_ratios=[1, 1, 2], subplot_spec=gs1[1])

        self.triangle = TrianglePlot(self.fig, gs0[0])
        self.temperature = Temperature(self.fig, gs0[1], self.update)
        self.evolution = Evolution(self.fig, gs10[0],
                                   self.update, len(self.run))
        self.higson = Higson(self.fig, gs10[1])
        self.reset = Button(self.fig, gs11[0],
                            self.reset_range, 'Reset Range')
        self.reload = Button(self.fig, gs11[1],
                             self.reload_file, 'Reload File')
        self.type = RadioButtons(self.fig, gs11[2],
                                 ('live', 'posterior'), self.update)
        self.param_choice = CheckButtons(self.fig, gs1[2],
                                         self.labels, self.redraw)

    def redraw(self, _):
        """ Redraw the triangle plot upon parameter updating."""
        self.triangle.draw(self.param_choice())
        self.update(None)
        self.reset_range(None)
        self.fig.tight_layout()
        self.fig.canvas.draw()

    def points(self, label):
        """ Get sample coordinates from nested sampling run.

        Args:
            label (str):
                label indicating the coordinate to extract.
        Returns:
            array(float):
                sample 'label'-coordinates.
        """
        if self.type() == 'posterior':
            kT = self.temperature()
            return self.run.posterior_points(1/kT)[label]
        else:
            i = self.evolution()
            logL = self.run.logL.iloc[i]
            return self.run.live_points(logL)[label]

    def update(self, _):
        """ Update all the plots upon slider changes."""
        logX = numpy.log(self.run.nlive/(self.run.nlive+1)).cumsum()
        kT = self.temperature()
        LX = self.run.logL/kT + logX
        LX = numpy.exp(LX-LX.max())
        i = self.evolution()
        logL = self.run.logL.iloc[i]
        try:
            n = self.run.nlive.iloc[i]
        except IndexError:
            n = 0

        self.triangle.update(self.points)

        self.evolution.set_text(logL, n)
        self.temperature.set_text(kT)

        self.higson.update(logX, LX, i)
        self.fig.canvas.draw()

    def reload_file(self, _):
        """ Reload the data from file."""
        self.run.reload_data()
        self.evolution.reset_range(valmax=len(self.run))
        self.update(None)

    def reset_range(self, _):
        """ Reset the parameter ranges. """
        self.triangle.reset_range()
        self.higson.reset_range()
        self.fig.canvas.draw()



import numpy
import pandas
from scipy.special import logsumexp
from anesthetic.plot import make_1D_axes, make_2D_axes, plot_1d, scatter_plot_2d, contour_plot_2d
from anesthetic.read import read_chains, read_birth, read_limits, read_paramnames
from anesthetic.information_theory import compress_weights


class MCMCSamples(pandas.DataFrame):
    """Storage and plotting tools for MCMC samples

    We extend the pandas.DataFrame by providing plotting methods and standardising
    sample storage.

    Note that because of the design of pandas this does not override the
    __init__ constructor. You should build the samples with either:

    * `mcmc = MCMCSamples.read('your/file/root')`
    * `mcmc = MCMCSamples.build(params=params, other_keyword_arguments)`

    Example plotting commands include

    * `mcmc.plot_1d()`
    * `mcmc.plot_2d(['paramA', 'paramB'])`
    * `mcmc.plot_2d(['paramA', 'paramB'],['paramC', 'paramD'])`

    """
    _metadata = pandas.DataFrame._metadata + ['paramnames', 'tex', 'limits', 'root']

    @classmethod
    def read(cls, root):
        """Read in data from file root."""
        # Read in data
        w, logL, params = read_chains(root)
        paramnames, tex = read_paramnames(root)
        limits = read_limits(root)

        # Build class
        data = cls.build(params=params, w=w, logL=logL, paramnames=paramnames,
                         tex=tex, limits=limits)

        # Record root
        data.root = root
        return data

    @classmethod
    def build(cls, **kwargs):
        """Build an augmented pandas array for MCMC samples.

        Parameters
        ----------
        params: numpy.array
            Coordinates of samples. shape = (nsamples, ndims).

        logL: numpy.array
            loglikelihoods of samples.

        w: numpy.array
            weights of samples.

        paramnames: list(str)
            reference names of parameters

        tex: dict
            mapping from paramnames to tex labels for plotting

        limits: dict
            mapping from paramnames to prior limits
        """
        params = kwargs.pop('params', None)
        logL = kwargs.pop('logL', None) 
        if params is None and logL is None:
            raise ValueError("You must provide either params or logL")
        elif params is None:
            params = numpy.empty((len(logL),0))

        nsamps, nparams = numpy.atleast_2d(params).shape

        w = kwargs.pop('w', None)
        paramnames = kwargs.pop('paramnames', ['x%i' % i for i in range(nparams)])

        tex = kwargs.pop('tex', {})
        limits = kwargs.pop('limits', {})

        data = cls(data=params, columns=paramnames)
        if w is not None:
            data['w'] = w 
            tex['w'] = r'MCMC weight'
        if logL is not None:
            data['logL'] = logL
            tex['logL'] = r'$\log\mathcal{L}$'

        data['u'] = numpy.random.rand(len(data))

        data.tex = tex
        data.paramnames = paramnames
        data.limits = limits
        data.root = None
        return data

    def plot(self, ax, paramname_x, paramname_y=None, *args, **kwargs):
        """Generic plotting interface. 
        
        Produces a single 1D or 2D plot on an axis.

        Parameters
        ----------
        ax: matplotlib.axes.Axes
            Axes to plot on 

        paramname_x: str
            Choice of parameter to plot on x-coordinate from self.columns.

        paramname_y: str
            Choice of parameter to plot on y-coordinate from self.columns.
            If not provided, or the same as paramname_x, then 1D plot produced.

        plot_type: str
            Optional. must be in {'contour','scatter'}

        beta: float
            Temperature to plot at. beta=0 corresponds to the prior, beta=1
            corresponds to the posterior.
        """

        plot_type = kwargs.pop('plot_type', 'contour')
        beta = kwargs.pop('beta', 1)

        if beta != 1 and isinstance(self, MCMCSamples):
            raise ValueError("You cannot adjust the temperature of MCMCSamples")

        if paramname_y is None or paramname_x == paramname_y:
            xmin, xmax = self._limits(paramname_x)
            return plot_1d(ax, numpy.repeat(self[paramname_x], self._weights(beta)),
                           xmin=xmin, xmax=xmax, *args, **kwargs)

        xmin, xmax = self._limits(paramname_x)
        ymin, ymax = self._limits(paramname_y)

        if plot_type == 'contour':
            weights = self._weights(beta)
            plot = contour_plot_2d
        elif plot_type == 'scatter':
            weights = self._weights(beta, nsamples=500)
            plot = scatter_plot_2d
        else:
            raise ValueError("plot_type must be in {'contour', 'scatter'}")

        return plot(ax, numpy.repeat(self[paramname_x], weights),
                    numpy.repeat(self[paramname_y], weights),
                    xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, 
                    *args, **kwargs)

    def plot_1d(self, paramnames=None, *args, **kwargs):
        """Create an array of 1D plots

        Parameters
        ----------
        paramnames: list(str) or str
            list of parameter names, or single parameter name to plot from
            self.columns. Optional, defaults to all parameters.

        axes: numpy.array(matplotlib.axes.Axes)
            Existing array of axes to plot on. If not provided, one is created.

        beta: float
            Temperature to plot at. beta=0 corresponds to the prior, beta=1
            corresponds to the posterior.

        Returns
        -------
        fig: matplotlib.figure.Figure
            New or original (if supplied) figure object

        axes: pandas.Series(matplotlib.axes.Axes)
            Pandas array of axes objects 
        """
        if paramnames is None:
            paramnames = self.paramnames
        else:
            paramnames = numpy.atleast_1d(paramnames)

        axes = kwargs.pop('axes', None)
        beta = kwargs.pop('beta', 1)
        if axes is None:
            fig, axes = make_1D_axes(paramnames, tex=self.tex)
        else:
            fig = numpy.atleast_2d(axes)[0,0].figure

        for p in paramnames:
            self.plot(axes[p], p, beta=beta, *args, **kwargs)

        return fig, axes

    def plot_2d(self, paramnames_x=None, paramnames_y=None, *args, **kwargs):
        """Create an array of 2D plots

        Parameters
        ----------
        paramnames_x: list(str) or str
            list of parameter names, or single parameter name to plot from
            self.columns. If paramnames_y is not provided, produce triangle plot

        paramnames_y: list(str) or str
            list of parameter names, or single parameter name to plot on y
            coordinate from self.columns. If not provided, then a triangle plot
            is produced from paramnames_x

        axes: numpy.array(matplotlib.axes.Axes)
            Existing array of axes to plot on. If not provided, one is created.

        beta: float
            Temperature to plot at. beta=0 corresponds to the prior, beta=1
            corresponds to the posterior.

        Returns
        -------
        fig: matplotlib.figure.Figure
            New or original (if supplied) figure object

        axes: pandas.DataFrame(matplotlib.axes.Axes)
            Pandas array of axes objects 
        """
        if paramnames_x is None:
            paramnames_x = self.paramnames
        paramnames_x = numpy.atleast_1d(paramnames_x)
        if paramnames_y is None:
            paramnames_y = paramnames_x
        paramnames_y = numpy.atleast_1d(paramnames_y)
        all_paramnames = list(paramnames_y) +list(paramnames_x)

        axes = kwargs.pop('axes', None)
        beta = kwargs.pop('beta', 1)
        if axes is None:
            fig, axes = make_2D_axes(paramnames_x, paramnames_y, tex=self.tex)
        else:
            axes = pandas.DataFrame(axes, index=paramnames_y, columns=paramnames_x)
            fig = axes.iloc[0,0].figure

        for py in paramnames_y:
            for px in paramnames_x:
                if px in paramnames_y and py in paramnames_x and all_paramnames.index(px) > all_paramnames.index(py):
                    plot_type='scatter'
                else:
                    plot_type='contour'
                self.plot(axes[px][py], px, py, plot_type=plot_type, beta=beta, *args, **kwargs)
        return fig, axes

    def _weights(self, beta, nsamples=None):
        """ Return the posterior weights for plotting. """
        try:
            return compress_weights(self.w, self.u, nsamples=nsamples)
        except AttributeError:
            return numpy.ones(len(self), dtype=int, unit_weights=unit_weights)

    def _limits(self, paramname):
        return self.limits.get(paramname, (None, None))

    def _reload_data(self):
        self = type(self).read(self.root)


class NestedSamples(MCMCSamples):
    """Storage and plotting tools for Nested Sampling samples

    We extend the MCMCSamples class with the additional methods:
    
    * ns_output

    Note that because of the design of pandas this does not override the
    __init__ constructor. You should build the samples with either:

    * NestedSamples.read('your/file/root')
    * NestedSamples.build(params=params, other_keyword_arguments)
    """

    @classmethod
    def read(cls, root):
        """Read in data from file root."""
        # Read in data
        paramnames, tex = read_paramnames(root)
        limits = read_limits(root)
        params, logL, logL_birth = read_birth(root)

        # Build class
        data = cls.build(params=params, logL=logL, paramnames=paramnames, tex=tex, limits=limits, logL_birth=logL_birth)

        # Record root
        data.root = root
        return data

    @classmethod
    def build(cls, **kwargs):
        """Build an augmented pandas array for Nested samples.

        Parameters
        ----------
        params: numpy.array
            Coordinates of samples. shape = (nsamples, ndims).

        logL: numpy.array
            loglikelihoods of samples.

        logL_birth: numpy.array
            birth loglikelihoods of samples.

        w: numpy.array
            weights of samples.

        paramnames: list(str)
            reference names of parameters

        tex: dict
            mapping from paramnames to tex labels for plotting

        limits: dict
            mapping from paramnames to prior limits
        """
        # Build pandas DataFrame
        logL_birth = kwargs.pop('logL_birth', None)
        data = super(NestedSamples, cls).build(**kwargs)
        data['logL_birth'] = logL_birth

        # Compute nlive
        index = data.logL.searchsorted(data.logL_birth)-1
        births = pandas.Series(+1, index=index).sort_index()
        deaths = pandas.Series(-1, index=data.index)
        nlive = pandas.concat([births, deaths]).sort_index().cumsum()
        nlive = (nlive[~nlive.index.duplicated(keep='first')]+1)[1:]
        data['nlive'] = nlive
        data['logw'] = data._dlogX()
        return data

    def ns_output(self, nsamples=100):
        """ Compute Bayesian global quantities

        Using nested sampling we can compute the evidence (logZ),
        Kullback-Leibler divergence (D) and Bayesian model dimensionality (d).
        More precisely, we can infer these quantities via their probability
        distribution.

        Parameters
        ----------
        nsamples: int
            number of samples to generate
            optional, default 100
        
        Returns
        -------
        pandas.DataFrame
            Samples from the P(logZ, D, d) distribution
        
        """
        columns = ['logZ', 'D', 'd']
        dlogX = self._dlogX(nsamples)

        logZ = logsumexp(self.logL.values + dlogX, axis=1)
        logw = ((self.logL.values + dlogX).T - logZ).T
        S = ((self.logL.values + numpy.zeros_like(dlogX)).T
             - logZ).T

        D = numpy.exp(logsumexp(logw, b=S, axis=1))
        d = numpy.exp(logsumexp(logw, b=(S.T-D).T**2, axis=1))*2

        params = numpy.vstack((logZ, D, d)).T
        paramnames = ['logZ', 'D', 'd']
        tex = {'logZ':r'$\log\mathcal{Z}$', 'D':r'$\mathcal{D}$', 'd':r'$d$'}
        return MCMCSamples.build(params=params, paramnames=paramnames, tex=tex)

    def live_points(self, logL):
        """ Get the live points within logL """
        return self[(self.logL > logL) & (self.logL_birth <= logL)]

    def posterior_points(self, beta):
        """ Get the posterior points at temperature beta """
        return self[self._weights(beta, nsamples=-1)>0]

    def _weights(self, beta, nsamples=None):
        """ Return the posterior weights for plotting. """
        logw = self.logw + beta*self.logL
        w = numpy.exp(logw - logw.max())
        return compress_weights(w, self.u, nsamples=nsamples)

    def _dlogX(self, nsamples=None):
        """ Compute volume of shell of loglikelihood

        Parameters
        ----------
        nsamples: int
            Number of samples to generate. optional. If None (default), then
            compute the statistical average. If integer, generate samples from
            the distribution
        """
        if nsamples is None:
            t = numpy.atleast_2d(numpy.log(self.nlive/(self.nlive+1)))
            nsamples=1
        else:
            t = numpy.log(numpy.random.rand(nsamples, len(self))
                          )/self.nlive.values
        logX = numpy.concatenate((numpy.zeros((nsamples, 1)),
                                  t.cumsum(axis=1),
                                  -numpy.inf*numpy.ones((nsamples, 1))
                                  ), axis=1)
        dlogX = logsumexp([logX[:, :-2], logX[:, 2:]],
                          b=[numpy.ones_like(t), -numpy.ones_like(t)], axis=0)
        dlogX -= numpy.log(2)
        return numpy.squeeze(dlogX)




import numpy

def check_bounds(d, xmin=None, xmax=None):
    if xmin is not None and (d.min() - xmin)/(d.max()-d.min()) >1e-2:
        xmin = None
    if xmax is not None and (xmax - d.max())/(d.max()-d.min()) >1e-2:
        xmax = None
    return xmin, xmax


def mirror_1d(d, xmin=None, xmax=None):
    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        return numpy.concatenate((2*xmin-d[d<xmed],d,2*xmax-d[d>=xmed]))
    elif xmin is not None:
        return numpy.concatenate((2*xmin-d,d))
    elif xmax is not None:
        return numpy.concatenate((d,2*xmax-d))
    else:
        return d


def mirror_2d(d_x_, d_y_, xmin=None, xmax=None, ymin=None, ymax=None):
    d_x = d_x_.copy()
    d_y = d_y_.copy()

    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        d_y = numpy.concatenate((d_y[d_x<xmed],d_y,d_y[d_x>=xmed]))
        d_x = numpy.concatenate((2*xmin-d_x[d_x<xmed],d_x,2*xmax-d_x[d_x>=xmed]))
    elif xmin is not None:
        d_y = numpy.concatenate((d_y,d_y))
        d_x = numpy.concatenate((2*xmin-d_x,d_x))
    elif xmax is not None:
        d_y = numpy.concatenate((d_y,d_y))
        d_x = numpy.concatenate((d_x,2*xmax-d_x))
        
    if ymin is not None and ymax is not None:
        ymed = (ymin+ymax)/2
        d_x = numpy.concatenate((d_x[d_y<ymed],d_x,d_x[d_y>=ymed]))
        d_y = numpy.concatenate((2*ymin-d_y[d_y<ymed],d_y,2*ymax-d_y[d_y>=ymed]))
    elif ymin is not None:
        d_x = numpy.concatenate((d_x,d_x))
        d_y = numpy.concatenate((2*ymin-d_y,d_y))
    elif ymax is not None:
        d_x = numpy.concatenate((d_x,d_x))
        d_y = numpy.concatenate((d_y,2*ymax-d_y))

    return d_x, d_y




import numpy
import warnings
from fastkde import fastKDE
from anesthetic.utils import check_bounds, mirror_1d, mirror_2d


def kde_1d(d, xmin=None, xmax=None):
    """ Perform a one-dimensional kernel density estimation

    Wrapper round fastkde.fastKDE. Boundary corrections implemented by
    reflecting boundary conditions.

    Parameters
    ----------
    d: numpy.array
        Data to perform kde on

    xmin, xmax: float
        lower/upper prior bounds 
        optional, default None
    
    Returns
    -------
    x: numpy.array
        x-coordinates of kernel density estimates
    p: numpy.array
        kernel density estimates
    """
    xmin, xmax = check_bounds(d, xmin, xmax)
    axisExpansionFactor = xmax is None or xmin is None
    d_ = mirror_1d(d, xmin, xmax)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, x = fastKDE.pdf(d_,axisExpansionFactor=axisExpansionFactor,
                           numPointsPerSigma=10*(2-axisExpansionFactor))
    p *= 2-axisExpansionFactor

    if xmin is not None:
        p = p[x>=xmin]
        x = x[x>=xmin]

    if xmax is not None:
        p = p[x<=xmax]
        x = x[x<=xmax]

    return x, p


def kde_2d(d_x, d_y, xmin=None, xmax=None, ymin=None, ymax=None):
    """ Perform a two-dimensional kernel density estimation

    Wrapper round fastkde.fastKDE. Boundary corrections implemented by
    reflecting boundary conditions.

    Parameters
    ----------
    d_x, d_y: numpy.array
        x/y coordinates of data to perform kde on

    xmin, xmax, ymin, ymax: float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    x,y: numpy.array
        x/y-coordinates of kernel density estimates. One-dimensional array
    p: numpy.array
        kernel density estimates. Two-dimensional array
    """
    xmin, xmax = check_bounds(d_x, xmin, xmax)
    ymin, ymax = check_bounds(d_y, ymin, ymax)
    axisExpansionFactor=[xmax is None or xmin is None,
                         ymax is None or ymin is None] 
    d_x_, d_y_ = mirror_2d(d_x, d_y, xmin, xmax, ymin, ymax)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, (x, y) = fastKDE.pdf(d_x_, d_y_, axisExpansionFactor=axisExpansionFactor,
                                numPointsPerSigma=10*(2-axisExpansionFactor[0])*(2-axisExpansionFactor[1]))

    p *= (2-axisExpansionFactor[0])
    p *= (2-axisExpansionFactor[1])
    if xmin is not None:
        p = p[:,x>=xmin]
        x = x[x>=xmin]

    if xmax is not None:
        p = p[:,x<=xmax]
        x = x[x<=xmax]

    if ymin is not None:
        p = p[y>=ymin,:]
        y = y[y>=ymin]

    if ymax is not None:
        p = p[y<=ymax,:]
        y = y[y<=ymax]

    return x, y, p



""" Widget wrappers to matplotlib.

These basically extend the matplotlib widgets by plotting themselves onto an
axis and storing a reference to both the widget object and the axis on which
they are plotted.

"""

from matplotlib.widgets import Button as mplButton
from matplotlib.widgets import CheckButtons as mplCheckButtons
from matplotlib.widgets import RadioButtons as mplRadioButtons
from rhinestone._matplotlib import Slider as mplSlider
from rhinestone.plot_utils import corner_plot, histogram
from rhinestone.utils import min_max


class Widget(object):
    """ Base class for rhinestone widgets widget.

    Stores a reference to the underlying figure, the gridspec that the
    widget is placed at and the axes of the widget.

    Args:
        fig (matplotlib.figure.Figure):
            Figure for drawing widget on.

        gridspec (matplotlib.gridspec.GridSpec):
            Specification for where to draw in the figure.
            Technically could be any argument that can be passed to
            matplotlib.figure.Figure.add_subplot.

    Attributes:
        fig  (matplotlib.figure.Figure):
            Figure for drawing widget on.

        gridspec (matplotlib.gridspec.GridSpec)
            Specification for where to draw in the figure.
            Technically could be any argument that can be passed to
            matplotlib.figure.Figure.add_subplot.

        ax  (matplotlib.axes.Axes):
            Axes of widget.
    """
    def __init__(self, fig, gridspec):
        self.fig = fig
        self.gridspec = gridspec
        self.ax = self.fig.add_subplot(self.gridspec)


class LabelsWidget(Widget):
    """ Widget with labels to choose from.

    Args:
        labels (list(str)):
            Set of labels to be tied to Widget.

    Attributes:
        labels (list(str)):
            Set of labels on Widget.
    """
    def __init__(self, fig, gridspec, labels):
        super(LabelsWidget, self).__init__(fig, gridspec)
        self.labels = labels


class Button(Widget):
    """ Push button that performs an action.

    Args:
        action (func):
            What should be run upon clicking the button.

        text (str):
            Overlay text on the button.

    Attributes:
        button (matplotlib.widgets.Button):
            matplotlib Button.
    """
    def __init__(self, fig, gridspec, action, text):
        super(Button, self).__init__(fig, gridspec)
        self.button = mplButton(self.ax, text)
        self.button.on_clicked(action)


class CheckButtons(LabelsWidget):
    """ Set of checkboxes.

    Defaults to choosing the only the first label at start.

    Args:
        action (func):
            What should be done upon checking the box.

    Attributes:
        buttons (matplotlib.widgets.CheckButtons):
            matplotlib CheckButtons.
    """
    def __init__(self, fig, gridspec, labels, action):
        super(CheckButtons, self).__init__(fig, gridspec, labels)
        default = [i == 0 for i, _ in enumerate(self.labels)]
        self.buttons = mplCheckButtons(self.ax, self.labels, default)
        self.buttons.on_clicked(action)

    def __call__(self):
        """ Current status of the check boxes.

        Returns:
            list(str):
                Labels that are currently selected.
        """
        return self.labels[list(self.buttons.get_status())]


class RadioButtons(LabelsWidget):
    """ Set of radio selection choices.

    Args:
        labels (list(str))

    Attributes:
        buttons (matplotlib.widgets.RadioButtons):
            matplotlib RadioButtons.
    """
    def __init__(self, fig, gridspec, labels, action):
        super(RadioButtons, self).__init__(fig, gridspec, labels)
        self.buttons = mplRadioButtons(self.ax, self.labels)
        self.buttons.on_clicked(action)

    def __call__(self):
        """ Current status of the radio selection.

        Returns:
            str:
                Label that is currently selected.
        """
        return self.buttons.value_selected


class Slider(Widget):
    """ Choose a parameter via a slider widget.

    Args:
        action (func):
            What should be done upon altering the slider.

        text (str):
            Text at the start of the slider.

        valmin, valmax, valinit (float):
            Range and initial value for the slider.

        orientation (str):
            Orientation for slider ('horizontal' or 'vertical').

    Attributes:
        slider (matplotlib.widgets.Slider):
            matplotlib Slider.
    """
    def __init__(self, fig, gridspec, action, text,
                 valmin, valmax, valinit, orientation):
        super(Slider, self).__init__(fig, gridspec)
        self.slider = mplSlider(self.ax, text, valmin, valmax, valinit=valinit,
                                orientation=orientation)
        self.slider.on_changed(action)

    def set_text(self, text):
        """ Set the text at the end of the slider.

        Args:
            text (str):
                Text to be chosen
        """
        self.slider.valtext.set_text(text)

    def reset_range(self, valmin=None, valmax=None):
        """ Reset the range of the slider

        Kwargs:
            valmin, valmax, (float), optional:
                The new limits to the slider.
        """
        if valmax is not None:
            self.slider.valmax = valmax
        if valmin is not None:
            self.slider.valmin = valmin
        self.slider.ax.set_xlim(self.slider.valmin, self.slider.valmax)

    def __call__(self):
        """ Return the float value in the slider."""
        return self.slider.val


class TrianglePlot(Widget):
    """ Triangle plot widget as exemplified by getdist and corner.py

    For other examples of these plots, see:
    https://getdist.readthedocs.io
    https://corner.readthedocs.io

    Attributes:
        ax (dict{(str, str):matplotlib.axes.Axes}):
            Mapping from pairs of parameters to axes for plotting.
            string pair is in x-y order.

        points (dict{(str, str):matplotlib.lines.line2D}):
            Mapping from pairs of parameters to points plotted on each axis.
            string pair is in x-y order.
    """
    def __init__(self, fig, gridspec):
        super(TrianglePlot, self).__init__(fig, gridspec)
        self.fig.delaxes(self.ax)
        self.ax = {}
        self.points = {}

    def draw(self, labels):
        """ Draw a new triangular grid for list of parameters labels.

        Args:
            labels (list(str)):
                labels for the triangular grid.
        """
        # Remove any existing axes
        for ax in self.ax.values():
            self.fig.delaxes(ax)

        # Set up the axes
        self.ax = corner_plot(self.fig, labels, self.gridspec)

        # Plot no points  points.
        self.points.clear()
        for (i, j), ax in self.ax.items():
            if i == j:
                fmt = 'k-'
            else:
                fmt = 'k.'
            self.points[(i, j)], = ax.plot([None], [None], fmt)

    def update(self, f):
        """ Update the points in the triangle plot using f function.

        Args:
            f (func: str -> array(float)):
                this function should take in a parameter label i, and return an
                array-like object of the i-coordinate of the samples
        """
        for (i, j), pts in self.points.items():
            if i == j:
                xdata, ydata = histogram(f(i), bins='auto')
                pts.set_xdata(xdata)
                pts.set_ydata(ydata)
            else:
                pts.set_xdata(f(i))
                pts.set_ydata(f(j))

    def reset_range(self):
        """ Reset the range of each grid. """
        for (i, j) in self.points:
            pts = self.points[(i, j)]
            ax = self.ax[(i, j)]
            xmin, xmax = min_max(pts.get_xdata())
            dx = xmax - xmin
            ax.set_xlim(xmin - dx*0.1, xmax + dx*0.1)
            if i != j:
                ymin, ymax = min_max(pts.get_ydata())
                dy = ymax - ymin
                ax.set_ylim(ymin - dy*0.1, ymax + dy*0.1)



import numpy

def read_paramnames(root):
    try:
        paramnames = []
        tex = {}
        paramnames_file = root + '.paramnames'
        with open(paramnames_file, 'r') as f:
            for line in f:
                line = line.strip().split()
                paramname = line[0].replace('*', '')
                paramnames.append(paramname)
                if len(line) > 1:
                    tex[paramname] = '$' + ''.join(line[1:]) + '$'
        return paramnames, tex
    except IOError:
        return None, None

def read_limits(root):
    limits_file = root + '.ranges'
    limits = {}
    try:
        with open(limits_file, 'r') as f:
            for line in f:
                line = line.strip().split()
                paramname = line[0]
                try:
                    xmin = float(line[1])
                except ValueError:
                    xmin = None
                try:
                    xmax = float(line[2])
                except ValueError:
                    xmax = None
                limits[paramname] = (xmin, xmax)
    except IOError:
        pass

    return limits


def read_birth(root):
    birth_file = root + '_dead-birth.txt'
    data = numpy.loadtxt(birth_file)
    params, logL, logL_birth  = numpy.split(data,[-2,-1], axis=1)
    return params, logL, logL_birth


def read_chains(root):
    chains_file = root + '_1.txt'
    data = numpy.loadtxt(chains_file)
    weights, chi2, params = numpy.split(data,[1,2], axis=1)
    logL = chi2/-2
    return weights, logL, params



import numpy
import pandas
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from fastkde import fastKDE
from anesthetic.kde import kde_1d, kde_2d
from anesthetic.utils import check_bounds
from scipy.interpolate import interp1d
from matplotlib.ticker import MaxNLocator
import matplotlib.colors

def make_1D_axes(paramnames, **kwargs):
    """ Create a set of axes for plotting 1D marginalised posteriors

    Parameters
    ----------
        paramnames: list(str)
            names of parameters.

        tex: dict(str:str)
            Dictionary mapping paramnames to tex plot labels.
            optional, default paramnames

        fig: matplotlib.figure.Figure
            Figure to plot on
            optional, default last figure matplotlib.pyplot.gcf()

        ncols: int
            Number of columns in the plot
            option, default ceil(sqrt(num_params))
            
        subplot_spec: matplotlib.gridspec.GridSpec
            gridspec to plot array as part of a subfigure
            optional, default None

    Returns
    -------
    fig: matplotlib.figure.Figure
        New or original (if supplied) figure object

    axes: pandas.Series(matplotlib.axes.Axes)
        Pandas array of axes objects 
    """

    tex = kwargs.pop('tex', {})
    fig = kwargs.pop('fig', plt.gcf())
    ncols = kwargs.pop('ncols', int(numpy.ceil(numpy.sqrt(len(paramnames)))))
    nrows = int(numpy.ceil(len(paramnames)/ncols))
    if 'subplot_spec' in kwargs:
        grid = gs.GridSpecFromSubplotSpec(nrows, ncols, wspace=0,
                                          subplot_spec=kwargs.pop('subplot_spec'))
    else:
        grid = gs.GridSpec(nrows, ncols, wspace=0)

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    tex = {p:tex[p] if p in tex else p for p in paramnames}
    axes = pandas.Series(index=paramnames, dtype=object)

    for p, g in zip(paramnames, grid):
        axes[p] = ax = fig.add_subplot(g) 
        ax.set_xlabel(tex[p])
        ax.set_yticks([])
        ax.xaxis.set_major_locator(MaxNLocator(2))

    return fig, axes


def make_2D_axes(paramnames, paramnames_y=None, **kwargs):
    """ Create a set of axes for plotting 2D marginalised posteriors

    Parameters
    ----------
        paramnames: list(str)
            names of parameters.

        paramnames_y: list(str)
            names of parameters.
            optional, default paramnames

        tex: dict(str:str)
            Dictionary mapping paramnames to tex plot labels.
            optional, default paramnames

        fig: matplotlib.figure.Figure
            Figure to plot on
            optional, default last figure matplotlib.pyplot.gcf()
            
        subplot_spec: matplotlib.gridspec.GridSpec
            gridspec to plot array as part of a subfigure
            optional, default None

    Returns
    -------
    fig: matplotlib.figure.Figure
        New or original (if supplied) figure object

    axes: pandas.DataFrame(matplotlib.axes.Axes)
        Pandas array of axes objects 
    """
    paramnames_x = paramnames
    if paramnames_y is None:
        paramnames_y = paramnames
    tex = kwargs.pop('tex', {})
    fig = kwargs.pop('fig', plt.gcf())
    nx = len(paramnames_x)
    ny = len(paramnames_y)
    if 'subplot_spec' in kwargs:
        grid = gs.GridSpecFromSubplotSpec(ny, nx, hspace=0, wspace=0,
                                          subplot_spec=kwargs.pop('subplot_spec'))
    else:
        grid = gs.GridSpec(ny, nx, hspace=0, wspace=0)

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    tex = {p:tex[p] if p in tex else p for p in numpy.concatenate((paramnames_x, paramnames_y))}
    axes = pandas.DataFrame(index=paramnames_y, columns=paramnames_x, dtype=object)
    axes[:][:] = None


    # 2D plots
    for y, py in enumerate(paramnames_y):
        for x, px in enumerate(paramnames_x):
            sx = axes[px][paramnames_y[0]]
            sy = axes[paramnames_y[0]][py]
            axes[px][py] = ax = fig.add_subplot(grid[y,x], sharex=sx, sharey=sy)
            
            ax.label_outer()

            if y == ny-1:
                ax.set_xlabel(tex[px])
                ax.xaxis.set_major_locator(MaxNLocator(2))
            else:
                ax.tick_params('x',bottom=False)

            if x == 0:
                ax.set_ylabel(tex[py])
                ax.yaxis.set_major_locator(MaxNLocator(2))
            else:
                ax.tick_params('y',left=False)

    return fig, axes


def plot_1d(ax, data, *args, **kwargs):
    """Plot a 1d marginalised distribution

    This functions as a wrapper around matplotlib.axes.Axes.plot, with a kernel
    density estimation computation in between. All remaining keyword arguments
    are passed onwards.

    To avoid intefering with y-axis sharing, one-dimensional plots are created
    on a separate axis, which is monkey-patched onto the argument ax as the
    attribute ax.twin.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data: numpy.array
        Uniformly weighted samples to generate kernel density estimator.

    xmin, xmax: float
        lower/upper prior bound
        optional, default None

    Returns
    -------
    lines: matplotlib.lines.Line2D
        A list of line objects representing the plotted data (same as
        matplotlib matplotlib.axes.Axes.plot command)
    """

    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)

    if not hasattr(ax, 'twin'):
        ax.twin = ax.twinx() 
        ax.twin.set_yticks([])
        ax.twin.set_ylim(0,1.1)
        if not ax.is_first_col():
            ax.tick_params('y',left=False)

    x, p = kde_1d(data, xmin, xmax)
    p /= p.max()
    i = (p>=1e-2)

    ans = ax.twin.plot(x[i], p[i], *args, **kwargs)
    ax.twin.set_xlim(*check_bounds(x[i], xmin, xmax), auto=True)
    return ans


def contour_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot a 2d marginalised distribution as contours

    This functions as a wrapper around matplotlib.axes.Axes.contour, and
    matplotlib.axes.Axes.contourf with a kernel density estimation computation
    in between. All remaining keyword arguments are passed onwards to both
    functions.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data_x, data_y: numpy.array
        x and y coordinates of uniformly weighted samples to generate kernel
        density estimator.

    xmin, xmax, ymin, ymax: float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    c: matplotlib.contour.QuadContourSet
        A set of contourlines or filled regions
    """

    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    ymin = kwargs.pop('ymin', None)
    ymax = kwargs.pop('ymax', None)
    color = kwargs.pop('color', next(ax._get_lines.prop_cycler)['color'])

    x, y, pdf = kde_2d(data_x, data_y, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
    pdf /= pdf.max()
    p = sorted(pdf.flatten())
    m = numpy.cumsum(p)
    m /= m[-1]
    interp = interp1d([0]+list(m)+[1],[0]+list(p)+[1])
    contours = list(interp([0.05, 0.33]))+[1]

    # Correct non-zero edges
    if min(p) != 0:
        contours = [min(p)] + contours

    # Correct level sets
    for i in range(1, len(contours)):
        if contours[i-1]==contours[i]:
            for j in range(i):
                contours[j] = contours[j] - 1e-5
            
    i = (pdf>=1e-2).any(axis=0)
    j = (pdf>=1e-2).any(axis=1)

    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(color, ['#ffffff', color])

    cbar = ax.contourf(x[i], y[j], pdf[numpy.ix_(j,i)], contours, vmin=0, vmax=1.0, cmap=cmap, *args, **kwargs)  
    ax.contour(x[i], y[j], pdf[numpy.ix_(j,i)], contours, vmin=0,vmax=1.2, linewidths=0.5, colors='k', *args, **kwargs)  
    ax.set_xlim(*check_bounds(x[i], xmin, xmax), auto=True)
    ax.set_ylim(*check_bounds(y[j], ymin, ymax), auto=True)
    return cbar


def scatter_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot samples from a 2d marginalised distribution

    This functions as a wrapper around matplotlib.axes.Axes.plot, enforcing any
    prior bounds. All remaining keyword arguments are passed onwards.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data_x, data_y: numpy.array
        x and y coordinates of uniformly weighted samples to generate kernel
        density estimator.

    xmin, xmax, ymin, ymax: float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    lines: matplotlib.lines.Line2D
        A list of line objects representing the plotted data (same as
        matplotlib matplotlib.axes.Axes.plot command)
    """
    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    ymin = kwargs.pop('ymin', None)
    ymax = kwargs.pop('ymax', None)

    points = ax.plot(data_x, data_y, 'o', markersize=1, *args, **kwargs)
    ax.set_xlim(*check_bounds(data_x, xmin, xmax), auto=True)
    ax.set_ylim(*check_bounds(data_y, ymin, ymax), auto=True)
    return points



#!/usr/bin/env python3
from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


def get_version(short=False):
    with open('README.rst') as f:
        for line in f:
            if ':Version:' in line:
                ver = line.split(':')[2].strip()
                if short:
                    subver = ver.split('.')
                    return '%s.%s' % tuple(subver[:2])
                else:
                    return ver


setup(name='anesthetic',
      version=get_version(),
      description='anesthetic: nested sampling post-processing',
      long_description=readme(),
      author='Will Handley',
      author_email='wh260@cam.ac.uk',
      url='https://github.com/williamjameshandley/anesthetic',
      packages=['anesthetic', 'rhinestone'],
      scripts=['scripts/rhinestone'],
      install_requires=['numpy', 'scipy', 'pandas', 'matplotlib', 'fastkde'],
      setup_requires=['pytest-runner'],
      extras_require={
          'docs': ['sphinx', 'sphinx_rtd_theme', 'numpydoc'],
          },
      tests_require=['pytest'],
      include_package_data=True,
      license='MIT',
      classifiers=[
                   'Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'Natural Language :: English',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Scientific/Engineering :: Visualization',
                   'Topic :: Scientific/Engineering :: Information Analysis',
                   'Topic :: Scientific/Engineering :: Mathematics',
      ],
      )



import numpy

def channel_capacity(w):
    with numpy.errstate(divide='ignore'):
        return numpy.exp(numpy.nansum((numpy.log(sum(w))-numpy.log(w))*w)/sum(w))


def compress_weights(w, u=None, nsamples=None):
    """ Compresses weights to their approximate channel capacity."""
    if u is None:
        u = numpy.random.rand(len(w))

    if nsamples is not None:
        W = w/w.max()
        if nsamples>0 and sum(W) > nsamples:
            W = W/sum(W)*nsamples
    else:
        W = w * channel_capacity(w) / w.sum()

    fraction, integer = numpy.modf(W)
    extra = (u<fraction).astype(int)
    return integer + extra



""" General plotting utilities."""
import numpy
import matplotlib.gridspec as gs


def corner_plot(fig, params, gridspec=None):
    """ Create a corner plot, as exemplified by getdist and corner.py

    For other examples of these plots, see:
    https://getdist.readthedocs.io
    https://corner.readthedocs.io

    Args:
        params (list(str)):
            list of labels for axes (best put in LaTeX)

    Kwargs:
        gridspec (matplotlib.gridspec.GridSpec):
            Specification for location of subplots in terms of
            matplotlib.gridspec.GridSpec.

    Returns:
        axes (dict{(str, str):matplotlib.axes.Axes}):
            Mapping from pairs of parameters to axes for plotting.
            string pair is in x-y order.
    """
    n = len(params)
    if gridspec is None:
        gridspec = gs.GridSpec(n, n)
    else:
        gridspec = gs.GridSpecFromSubplotSpec(n, n, subplot_spec=gridspec)

    # Set up the diagonal axes
    axes = {}

    # Set up the axes
    for j, p_y in enumerate(params):
        for i, p_x in enumerate(params):
            if i == j:
                axes[(p_x, p_y)] = fig.add_subplot(gridspec[j, i])
                axes[(p_x, p_y)].set_ylim(0, 1.1)
            if i < j:
                sx = axes[(p_x, p_x)]
                if i is 0:
                    sy = None
                else:
                    sy = axes[(params[0], p_y)]
                axes[(p_x, p_y)] = fig.add_subplot(gridspec[j, i],
                                                   sharex=sx, sharey=sy)

    # Do the labels
    for p in params:
        axes[(p, params[-1])].set_xlabel(p)
        if p != params[0]:
            axes[(params[0], p)].set_ylabel(p)

    # Remove unecessary ticks on top axis
    for p in params[:1]:
        axes[(p, p)].set_yticks([])
    for ax in axes.values():
        ax.label_outer()

    return axes


def histogram(a, **kwargs):
    """ Produce a histogram for path-based plotting

    This is a cheap histogram. Necessary if one wants to update the histogram
    dynamically, and redrawing and filling is very expensive.

    This has the same arguments and keywords as numpy.histogram, but is
    normalised to 1.
    """
    hist, bin_edges = numpy.histogram(a, **kwargs)
    xpath, ypath = numpy.empty((2, 4*len(hist)))
    ypath[0::4] = ypath[3::4] = 0
    ypath[1::4] = ypath[2::4] = hist
    xpath[0::4] = xpath[1::4] = bin_edges[:-1]
    xpath[2::4] = xpath[3::4] = bin_edges[1:]
    mx = max(ypath)
    if mx:
        ypath /= max(ypath)
    return xpath, ypath



"""General utilities."""


def min_max(x):
    """ Get the minimum and maximum of x at the same time.

    Args:
        x (array-like):
            anything that can be acted on by min() and max() functions.

    Returns:
        tuple:
            min and max of x
    """
    return min(x), max(x)
